﻿using PostSharp.Aspects.Configuration;
using PostSharp.Extensibility;
using PostSharp.Sdk.AspectWeaver;

namespace FastTrace.Weaver
{
    internal class FastTraceAspectWeaver : AspectWeaver
    {
        private static readonly AspectConfigurationAttribute defaultConfiguration = new AspectConfigurationAttribute();
        private FastTraceTransformation transformation;

        public FastTraceAspectWeaver()
            : base(
                defaultConfiguration, ReflectionObjectBuilder.Dynamic,
                MulticastTargets.Method | MulticastTargets.InstanceConstructor | MulticastTargets.StaticConstructor | MulticastTargets.Field )
        {
            this.RequiresRuntimeInstance = false;
            this.RequiresRuntimeReflectionObject = false;
        }

        protected override void Initialize()
        {
            base.Initialize();

            this.transformation = new FastTraceTransformation( this );
            this.ApplyWaivedEffects( this.transformation );
        }

        protected override AspectWeaverInstance CreateAspectWeaverInstance( AspectInstanceInfo aspectInstanceInfo )
        {
            return new Instance( this, aspectInstanceInfo );
        }

        private class Instance : AspectWeaverInstance
        {
            public Instance( AspectWeaver aspectWeaver, AspectInstanceInfo aspectInstanceInfo ) : base( aspectWeaver, aspectInstanceInfo )
            {
            }

            public override void ProvideAspectTransformations( AspectWeaverTransformationAdder adder )
            {
                adder.Add( this.TargetElement, ((FastTraceAspectWeaver) this.AspectWeaver).transformation.CreateInstance( this ) );
            }
        }
    }
}