﻿using System;
using PostSharp.Sdk.AspectInfrastructure;
using PostSharp.Sdk.AspectWeaver;
using PostSharp.Sdk.AspectWeaver.Transformations;
using PostSharp.Sdk.CodeModel;
using PostSharp.Sdk.Collections;

namespace FastTrace.Weaver
{
    internal class FastTraceTransformation : MethodBodyTransformation
    {
        private readonly Assets assets;

        public FastTraceTransformation( AspectWeaver aspectWeaver ) : base( aspectWeaver )
        {
            ModuleDeclaration module = this.AspectInfrastructureTask.Project.Module;
            this.assets = module.Cache.GetItem( () => new Assets( module ) );
        }


        public override string GetDisplayName( MethodSemantics semantic )
        {
            return "FastTrace";
        }

        public AspectWeaverTransformationInstance CreateInstance( AspectWeaverInstance aspectWeaverInstance )
        {
            return new Instance( this, aspectWeaverInstance );
        }

        private sealed class Instance : MethodBodyTransformationInstance
        {
            public Instance( MethodBodyTransformation parent, AspectWeaverInstance aspectWeaverInstance ) : base( parent, aspectWeaverInstance )
            {
            }

            public override void Implement( MethodBodyTransformationContext context )
            {
                Implementation implementation = new Implementation( this, context );
                implementation.Implement();
            }

            public override MethodBodyTransformationOptions GetOptions( MetadataDeclaration originalTargetElement, MethodSemantics semantic )
            {
                return MethodBodyTransformationOptions.None;
            }


            private sealed class Implementation : MethodBodyWrappingImplementation
            {
                private readonly string codeElementName;
                private readonly Instance transformationInstance;

                public Implementation( Instance transformationInstance, MethodBodyTransformationContext context )
                    : base( transformationInstance.AspectWeaver.AspectInfrastructureTask, context )
                {
                    this.transformationInstance = transformationInstance;
                    this.codeElementName = string.Format( "{0} ({1})", transformationInstance.AspectWeaverInstance.TargetElement.ToString(),
                                                          context.MethodSemantic );
                }

                public void Implement()
                {
                    this.Implement( true, false, true, null );
                    this.Context.AddRedirection( this.Redirection );
                }

                protected override void ImplementOnException( InstructionBlock block, ITypeSignature exceptionType, InstructionWriter writer )
                {
                }

                protected override void ImplementOnExit( InstructionBlock block, InstructionWriter writer )
                {
                    this.EmitMessage( block, writer, "Exiting " + this.codeElementName );
                }

                protected override void ImplementOnSuccess( InstructionBlock block, InstructionWriter writer )
                {
                }

                protected override void ImplementOnEntry( InstructionBlock block, InstructionWriter writer )
                {
                    this.EmitMessage( block, writer, "Entering " + this.codeElementName );
                }

                private void EmitMessage( InstructionBlock block, InstructionWriter writer, string message )
                {
                    InstructionSequence sequence = block.AddInstructionSequence( null, NodePosition.After, null );
                    writer.AttachInstructionSequence( sequence );
                    writer.EmitInstructionString( OpCodeNumber.Ldstr, message );
                    writer.EmitInstructionMethod( OpCodeNumber.Call,
                                                  ((FastTraceTransformation) this.transformationInstance.Transformation).assets.WriteLineMethod );
                    writer.DetachInstructionSequence();
                }
            }
        }


        private sealed class Assets
        {
            public readonly IMethod WriteLineMethod;

            public Assets( ModuleDeclaration module )
            {
                this.WriteLineMethod = module.FindMethod( typeof(Console).GetMethod( "WriteLine", new[] {typeof(string)} ), BindingOptions.Default );
            }
        }
    }
}